/*
 * $Id: ConcurrentRequestMap.java 286 2005-12-03 01:07:16Z davidb $
 * 
 * Copyright (c) 2005 VeriSign. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 3. The name of the author may not
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package se.rfc.unbound;

import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

/**
 * This class represents a map used to track currently active requests for the
 * purposes of preventing redundant requests from processing at the same time.
 * Each module that is attempting to suppress duplicate/redundant requests
 * should have a separate instance of this class.
 * 
 * @author davidb
 * @version $Revision: 286 $
 */
public class ConcurrentRequestMap
{
  public final static int DUPLICATE = 0;
  public final static int ORIGINAL  = 1;
  public final static int DEADLOCK  = 2;

  private Map             mMap;

  private Logger          log       = Logger.getLogger(this.getClass());

  /**
   * Standard constructor.
   */
  public ConcurrentRequestMap()
  {
    mMap = new HashMap();
  }

  /**
   * Generate a hash key for the entry.
   * 
   * @param event The event to key.
   * @return A string based hash key.
   */
  private String key(DNSEvent event)
  {
    Request r = event.getRequest();
    return Util.nameToString(r.getQName()) + "/" + r.getQType() + "/"
        + r.getQClass();
  }

  /**
   * Add the given event to the concurrent-request map.
   * 
   * @param event The event to (potentially add).
   * @param module The current module (this is used to fetch the proper
   *               {@link ConcurrentRequestState} object from the given event).
   * @return A result of DUPLICATE if event was a duplicate and was attached,
   *         ORIGINAL if the event wasn't a duplicate and was simply added to
   *         the map, or DEADLOCK if the event is a duplicate, but attaching
   *         it would cause a deadlock (defined by a circular dependency).
   * 
   */
  public int add(DNSEvent event, Resolver module)
  {
    String k = key(event);

    DNSEvent ev = getMatchingEvent(event);
    if (ev == null)
    {
      mMap.put(k, event);
      return ORIGINAL;
    }
    // check to make sure ev isn't actually depending on event (which will
    // create a deadlock)
    if (isDependent(ev, event))
    {
      log.debug("did not attach event '" + event
          + "' to in progress event to avoid deadlock");
      return DEADLOCK;
    }

    ConcurrentRequestState state = (ConcurrentRequestState) ev
        .getModuleState(module);
    state.addConcurrentEvent(event);
    return ORIGINAL;
  }

  /**
   * Check to see if the event, if attached to ev, would create a circular
   * dependency. This would happen if event ultimately depended on itself.
   * 
   * @param ev The candidate event to attach to.
   * @param event The candidate event to attach.
   * @return true if event appears in the forEvent list of ev.
   */
  private boolean isDependent(DNSEvent ev, DNSEvent event)
  {
    DNSEvent cur_ev = event.forEvent();
    while (cur_ev != null)
    {
      if (cur_ev == ev) return true;
      cur_ev = cur_ev.forEvent();
    }

    return false;
  }

  /**
   * Find a currently in-progress event that matches the given event.
   * 
   * @param event The event to match.
   * @return A matching event if there is one, null otherwise.
   */
  public DNSEvent getMatchingEvent(DNSEvent event)
  {
    String k = key(event);
    return (DNSEvent) mMap.get(k);
  }

  /**
   * Remove an event from the map. This must be done when the event is no
   * longer in progress.
   * 
   * @param event
   */
  public void remove(DNSEvent event)
  {
    String k = key(event);
    mMap.remove(k);
  }
}
